@@ -51,6 +51,7 @@ import softwareproperties
 from AptAuth import AptAuth
 from aptsources.sourceslist import SourcesList, SourceEntry
 from ppa import AddPPASigningKeyThread, expand_ppa_line
+from cloudarchive import shortcut_handler as ca_shortcut_handler
 
 class SoftwareProperties(object):
 
@@ -420,6 +421,10 @@ class SoftwareProperties(object):
         self.sourceslist_backup.append(source_bkp)
   
   def _find_source_from_string(self, line):
+    # ensure that we have a current list, it might have been changed underneath
+    # us
+    self.reload_sourceslist()
+
     for source in self.sourceslist.list:
       if str(source) == line:
         return source
@@ -660,6 +665,9 @@ class SoftwareProperties(object):
     if parsed_uri.netloc == 'ppa.launchpad.net':
       worker = AddPPASigningKeyThread(parsed_uri.path, self.options and self.options.keyserver)
       worker.start()
+      return worker
+    else:
+      return None
 
   def update_interface(self):
     " abstract interface to keep the UI alive "
@@ -686,17 +694,41 @@ class SoftwareProperties(object):
     Add a source with the given apt line and auto-add
     signing key if we have it in the whitelist
     """
-    (deb_line, file) = expand_ppa_line(line.strip(), self.distro.codename)
+    (deb_line, file) = expand_shortcut_line(line.strip(), self.distro.codename)
     deb_line = self.expand_http_line(deb_line)
     debsrc_line = 'deb-src' + deb_line[3:]
     new_deb_entry = SourceEntry(deb_line, file)
     new_debsrc_entry = SourceEntry(debsrc_line, file)
     if new_deb_entry.invalid or new_debsrc_entry.invalid:
       return False
-    self.check_and_add_key_for_whitelisted_channels(deb_line)
-    self.sourceslist.list.append(new_deb_entry)
-    self.sourceslist.list.append(new_debsrc_entry)
+
+    if line.startswith("cloud-archive:"):
+        # in trunk, we modified check_and_add_key_for_whitelisted_channels to
+        # take the line itself.  but this is that functions only caller
+        # so it should be safe and less invasive to do this check here instead.
+        ca_handler = ca_shortcut_handler(line.strip())
+        worker = threading.Thread(target=ca_handler.add_key)
+        worker.start()
+    else:
+        worker = self.check_and_add_key_for_whitelisted_channels(deb_line)
+    self.sourceslist.add(new_deb_entry.type,
+                         new_deb_entry.uri,
+                         new_deb_entry.dist,
+                         new_deb_entry.comps,
+                         comment=new_deb_entry.comment,
+                         file=new_deb_entry.file,
+                         architectures=new_deb_entry.architectures)
+    self.sourceslist.add(new_debsrc_entry.type,
+                         new_debsrc_entry.uri,
+                         new_debsrc_entry.dist,
+                         new_debsrc_entry.comps,
+                         comment=new_debsrc_entry.comment,
+                         file=new_debsrc_entry.file,
+                         architectures=new_debsrc_entry.architectures)
     self.set_modified_sourceslist()
+    if worker:
+        # wait for GPG key to be downloaded
+        worker.join(30)
     return True
 
   def remove_source(self, source):
@@ -752,6 +784,14 @@ class SoftwareProperties(object):
       return False
 
 
+def expand_shortcut_line(abrev, codename):
+    if abrev.startswith("cloud-archive:"):
+        handler = ca_shortcut_handler(abrev)
+        return handler.expand(codename)
+
+    return expand_ppa_line(abrev, distro_codename=codename)
+
+
 if __name__ == "__main__":
   sp = SoftwareProperties()
   print sp.get_release_upgrades_policy()
